"use strict";
// cspell:words nothrow
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSpawn = void 0;
const Logger_1 = require("./Logger");
const throttle_debounce_1 = require("throttle-debounce");
const zx_1 = require("zx");
const log = Logger_1.Logger.child({
    namespace: 'createSpawn',
});
const prefixLines = (subject, prefix) => {
    const response = [];
    for (const fragment of subject.split('\n')) {
        response.push(prefix + fragment);
    }
    return response.join('\n');
};
const createSpawn = (taskId, { abortSignal, throttleOutput, } = {}) => {
    let stdoutBuffer = [];
    let stderrBuffer = [];
    const flush = () => {
        if (stdoutBuffer.length) {
            // eslint-disable-next-line no-console
            console.log(stdoutBuffer.join('\n'));
        }
        if (stderrBuffer.length) {
            // eslint-disable-next-line no-console
            console.error(stderrBuffer.join('\n'));
        }
        stdoutBuffer = [];
        stderrBuffer = [];
    };
    const output = (0, throttle_debounce_1.throttle)(throttleOutput === null || throttleOutput === void 0 ? void 0 : throttleOutput.delay, () => {
        flush();
    }, {
        noLeading: true,
    });
    return async (pieces, ...args) => {
        // eslint-disable-next-line promise/prefer-await-to-then
        const processPromise = (0, zx_1.$)(pieces, ...args)
            .nothrow()
            .quiet();
        (async () => {
            for await (const chunk of processPromise.stdout) {
                const message = prefixLines(chunk.toString().trimEnd(), taskId + ' > ');
                if (throttleOutput === null || throttleOutput === void 0 ? void 0 : throttleOutput.delay) {
                    stdoutBuffer.push(message);
                    output();
                }
                else {
                    // eslint-disable-next-line no-console
                    console.log(message);
                }
            }
        })();
        (async () => {
            for await (const chunk of processPromise.stderr) {
                const message = prefixLines(chunk.toString().trimEnd(), taskId + ' > ');
                if (throttleOutput === null || throttleOutput === void 0 ? void 0 : throttleOutput.delay) {
                    stderrBuffer.push(message);
                    output();
                }
                else {
                    // eslint-disable-next-line no-console
                    console.error(message);
                }
            }
        })();
        if (abortSignal) {
            const kill = () => {
                processPromise.kill();
            };
            abortSignal.addEventListener('abort', kill, {
                once: true,
            });
            // eslint-disable-next-line promise/prefer-await-to-then
            processPromise.finally(() => {
                abortSignal.removeEventListener('abort', kill);
            });
        }
        const result = await processPromise;
        flush();
        if (result.exitCode === 0) {
            return result;
        }
        if (abortSignal === null || abortSignal === void 0 ? void 0 : abortSignal.aborted) {
            return result;
        }
        log.error('task %s exited with an error', taskId);
        throw new Error('Program exited with code ' + result.exitCode + '.');
    };
};
exports.createSpawn = createSpawn;
//# sourceMappingURL=createSpawn.js.map