"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.subscribe = void 0;
const createSpawn_1 = require("./createSpawn");
const generateShortId_1 = require("./generateShortId");
const Logger_1 = require("./Logger");
const p_retry_1 = __importDefault(require("p-retry"));
const log = Logger_1.Logger.child({
    namespace: 'subscribe',
});
const subscribe = (trigger) => {
    let activeTask = null;
    let first = true;
    const handleSubscriptionEvent = async (event) => {
        var _a;
        if ((_a = trigger.abortSignal) === null || _a === void 0 ? void 0 : _a.aborted) {
            log.warn('ignoring event because Turbowatch is shutting down');
            return undefined;
        }
        if (trigger.initialRun && first) {
            log.trace('initial run...');
        }
        else if (event.files.length > 10) {
            log.trace({
                files: event.files.slice(0, 10).map((file) => {
                    return file.name;
                }),
            }, '%d files changed; showing first 10', event.files.length);
        }
        else {
            log.trace({
                files: event.files.map((file) => {
                    return file.name;
                }),
            }, '%d files changed', event.files.length);
        }
        let reportFirst = first;
        if (first) {
            reportFirst = true;
            first = false;
        }
        let controller = null;
        if (trigger.interruptible) {
            controller = new AbortController();
        }
        let abortSignal = controller === null || controller === void 0 ? void 0 : controller.signal;
        if (abortSignal && trigger.abortSignal) {
            trigger.abortSignal.addEventListener('abort', () => {
                controller === null || controller === void 0 ? void 0 : controller.abort();
            });
        }
        else if (trigger.abortSignal) {
            abortSignal = trigger.abortSignal;
        }
        if (activeTask) {
            if (trigger.interruptible) {
                log.warn('aborted task %s (%s)', trigger.name, activeTask.id);
                if (!activeTask.abortController) {
                    throw new Error('Expected abort controller to be set');
                }
                activeTask.abortController.abort();
                activeTask = null;
            }
            else {
                log.warn('waiting for %s (%s) task to complete', trigger.name, activeTask.id);
                if (activeTask.queued) {
                    return undefined;
                }
                activeTask.queued = true;
                try {
                    await activeTask.promise;
                }
                catch (_b) {
                    // nothing to do
                }
            }
        }
        const taskId = (0, generateShortId_1.generateShortId)();
        const taskPromise = (0, p_retry_1.default)((attempt) => {
            return trigger.onChange({
                abortSignal,
                attempt,
                files: event.files.map((file) => {
                    return {
                        name: file.name,
                    };
                }),
                first: reportFirst,
                spawn: (0, createSpawn_1.createSpawn)(taskId, {
                    abortSignal,
                    throttleOutput: trigger.throttleOutput,
                }),
                taskId,
            });
        }, {
            ...trigger.retry,
            onFailedAttempt: ({ retriesLeft }) => {
                if (retriesLeft > 0) {
                    log.warn('retrying task %s (%s)...', trigger.name, taskId);
                }
            },
        })
            // eslint-disable-next-line promise/prefer-await-to-then
            .then(() => {
            if (taskId === (activeTask === null || activeTask === void 0 ? void 0 : activeTask.id)) {
                log.trace('completed task %s (%s)', trigger.name, taskId);
                activeTask = null;
            }
        });
        // eslint-disable-next-line require-atomic-updates
        activeTask = {
            abortController: controller,
            id: taskId,
            promise: taskPromise,
            queued: false,
        };
        log.trace('started task %s (%s)', trigger.name, taskId);
        return taskPromise;
    };
    return {
        activeTask,
        expression: trigger.expression,
        initialRun: trigger.initialRun,
        teardown: async () => {
            if (trigger.onTeardown) {
                const taskId = (0, generateShortId_1.generateShortId)();
                await trigger.onTeardown({
                    spawn: (0, createSpawn_1.createSpawn)(taskId, {
                        throttleOutput: trigger.throttleOutput,
                    }),
                });
            }
        },
        trigger: async (events) => {
            await handleSubscriptionEvent({
                files: events.map((event) => {
                    return {
                        name: event.path,
                    };
                }),
            });
        },
    };
};
exports.subscribe = subscribe;
//# sourceMappingURL=subscribe.js.map